package info.textgrid.utils.export.filenames;

import info.textgrid.namespaces.metadata.core._2010.ObjectType;
import info.textgrid.utils.export.aggregations.IAggregationEntry;

import java.net.URI;
import java.text.CharacterIterator;
import java.text.StringCharacterIterator;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.google.common.base.Joiner;
import com.google.common.base.Optional;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableList.Builder;
import com.google.common.collect.Maps;
import com.ibm.icu.text.Transliterator;

public class ConfigurableFilenamePolicy implements IFilenamePolicy {

	public class SubPolicySegment extends MetadataSegment implements Segment {

		private ConfigurableFilenamePolicy policy;

		public SubPolicySegment(ConfigurableFilenamePolicy policy, String... options) {
			super(options);
			this.policy = policy; 
			this.raw = true;
		}

		@Override
		public String getContent(IAggregationEntry entry) {
			return policy.getFilename(entry, false).toString();
		}

	}

	private final Map<String, ConfigurableFilenamePolicy> subPolicies = Maps
			.newHashMap();

	/**
	 * Adds a sub policy to this policy. If the pattern contains the variable
	 * <var>name</var>, it will be expanded by calling the policy that is
	 * configured here.
	 * 
	 * @param name
	 * @param policy
	 * @return
	 */
	public ConfigurableFilenamePolicy addSubPolicy(final String name,
			final ConfigurableFilenamePolicy policy) {
		subPolicies.put(name, policy);
		return this;
	}

	public class ExtensionSegment extends MetadataSegment implements Segment {

		public ExtensionSegment(String... options) {
			super(options);
			raw = true; // extensions need never be transliterated.
		}

		@Override
		public String getContent(IAggregationEntry entry) {
			String format = entry.getMetadata().getGeneric().getProvided()
					.getFormat();
			return FileExtensionMap.getInstance().getFirstExtension(format)
					.or("dat");
		}

	}

	/** A segment represents a part of a filename pattern. */
	protected interface Segment {
		/**
		 * Appends the expansion of this segment for the given entry to the
		 * given builder.
		 */
		public StringBuilder append(final StringBuilder builder,
				IAggregationEntry entry);
	}

	/**
	 * A segment that always expands to a static string, regardless of the entry
	 */
	protected class StaticSegment implements Segment {
		protected final String content;

		/**
		 * Creates a static string segment.
		 * 
		 * @param content
		 *            The string this segment will expand to.
		 */
		public StaticSegment(String content) {
			this.content = content;
		}

		@Override
		public StringBuilder append(StringBuilder builder,
				IAggregationEntry entry) {
			return builder.append(content);
		}

	}

	/** Base class for segments that extract metadata from the given argument. */
	protected abstract class MetadataSegment implements Segment {
		/**
		 * Hint to insert the raw content of the metadata field, without any
		 * translation to safe characters.
		 */
		protected boolean raw = false;

		/**
		 * Hint that suggests the maximal number of characters to append for
		 * this segment. Values <= 0 mean no restriction.
		 */
		protected int maxlength = -1;

		/**
		 * Initializes this segment. Calls {@link #handleOption(String)} for
		 * each given option.
		 */
		public MetadataSegment(String... options) {
			for (String option : options)
				handleOption(option);
		}

		/**
		 * Parse the given option. Clients may override or extend.
		 * 
		 * @param option
		 *            The option. Currently supported strings are
		 *            <code>raw</code> to set the {@link #raw} flag or any
		 *            positive decimal integer to set the {@link #maxlength}
		 *            option.
		 * @return <code>true</code> iff the method handled the option
		 */
		protected boolean handleOption(String option) {
			if ("raw".equals(option))
				raw = true;
			else if (option.matches("\\d+"))
				maxlength = Integer.parseInt(option);
			else
				return false;
			return true;

		}

		/** Return the (unprocessed) string for this entry. */
		public abstract String getContent(IAggregationEntry entry);

		/**
		 * {@inheritDoc}
		 * 
		 * The {@link MetadataSegment} implementation calls
		 * {@link #getContent(IAggregationEntry)} and postprocesses the result
		 * according to the {@link #raw} and {@link #maxlength} fields before
		 * appending.
		 */
		public StringBuilder append(StringBuilder builder,
				IAggregationEntry entry) {
			String content = getContent(entry);
			if (!raw)
				content = translate(content);
			if (maxlength > 0)
				content = content.substring(0, maxlength - 1);
			return builder.append(content);
		}

	}

	/**
	 * Formats an object's TextGrid URI.
	 */
	public class URISegment extends MetadataSegment implements Segment {

		public URISegment(String... options) {
			super(options);
		}

		@Override
		public String getContent(IAggregationEntry entry) {
			return entry.getMetadata().getGeneric().getGenerated()
					.getTextgridUri().getValue();
		}

	}

	/**
	 * Extracts the title.
	 * 
	 * If the additional option sep=<string> is given and the given object has
	 * multiple
	 * 
	 */
	public class TitleSegment extends MetadataSegment {

		protected Optional<String> separator = Optional.absent();

		public TitleSegment(String... options) {
			super(options);
		}

		private final Pattern SEP = Pattern.compile("^sep=(.+)$");

		@Override
		protected boolean handleOption(String option) {
			if (super.handleOption(option))
				return true;
			else {
				Matcher matcher = SEP.matcher(option);
				if (matcher.matches()) {
					separator = Optional.of(matcher.group(1));
					return true;
				}
				return false;
			}
		}

		public List<String> getContentList(IAggregationEntry entry) {
			return entry.getMetadata().getGeneric().getProvided().getTitle();
		}

		@Override
		public String getContent(IAggregationEntry entry) {
			List<String> list = getContentList(entry);
			if (separator.isPresent())
				return Joiner.on(separator.get()).join(list);
			else if (list.size() > 0)
				return list.get(0);
			else
				return "";
		}

	}

	private final Transliterator transliterator;
	private ImmutableList<Segment> segments;

	public ConfigurableFilenamePolicy(final String pattern) {
		this.transliterator = DefaultFilenamePolicy.getTransliterator();

		StringBuilder buffer = new StringBuilder();
		Builder<Segment> segments = ImmutableList.builder();
		boolean escape = false;

		StringCharacterIterator iter = new StringCharacterIterator(pattern);

		for (char c = iter.first(); c != CharacterIterator.DONE; c = iter
				.next())
			if (escape) {
				buffer.append(c);
				escape = false;
			} else if (c == '\\')
				escape = true;
			else if (c == '{') {
				String s = buffer.toString();
				if (!s.isEmpty()) {
					segments.add(new StaticSegment(s));
					buffer = new StringBuilder();
				}
				segments.add(parseVariable(iter));
			} else
				buffer.append(c);
		this.segments = segments.build();
	}

	private Segment parseVariable(StringCharacterIterator iter) {
		if (iter.current() == '{')
			iter.next();
		StringBuilder buf = new StringBuilder();
		for (char c = iter.current(); c != '}' && c != CharacterIterator.DONE; c = iter
				.next())
			buf.append(c);
		String[] split = buf.toString().split(",");
		String var = split[0];
		String[] options = Arrays.copyOfRange(split, 1, split.length);
		
		if (subPolicies.containsKey(var)) 
			return new SubPolicySegment(subPolicies.get(var), options);
		else if (("uri").equals(var))
			return new URISegment(options);
		else if (("title".equals(var)))
			return new TitleSegment(options);
		else if (("ext".equals(var)))
			return new ExtensionSegment(options);
		return null;
	}

	@Override
	public String translate(String source) {
		return transliterator.transliterate(source);
	}

	@Override
	public String getFilename(ObjectType object, boolean asParent) {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public URI getFilename(IAggregationEntry entry, boolean asParent) {
		StringBuilder builder = new StringBuilder();
		for (Segment segment : segments)
			segment.append(builder, entry);
		return URI.create(builder.toString());
	}

	@Override
	public Optional<URI> getBase(IAggregationEntry entry) {
		// TODO Auto-generated method stub
		return null;
	}

}
